@@ -31,21 +31,10 @@ from typing import Any
 import os
 import sys
 import unicodedata
+import functools
 import logging
 import itb_util
 
-IMPORT_REGEX_SUCCESFUL = False
-try:
-    import regex # type: ignore
-    IMPORT_REGEX_SUCCESFUL = True
-    # Enable new improved regex engine instead of backwards compatible
-    # v0.  regex.match('ß', 'SS', regex.IGNORECASE) matches only with
-    # the improved version!  See also: https://pypi.org/project/regex/
-    regex.DEFAULT_VERSION = regex.VERSION1 # pylint: disable=no-member
-except (ImportError,):
-    # Use standard “re” module as a fallback:
-    import re
-
 LOGGER = logging.getLogger('ibus-typing-booster')
 
 DEBUG_LEVEL = int(0)
@@ -331,6 +320,20 @@ class Dictionary():
             self.voikko.suggest(unicodedata.normalize('NFC', word))
             ]
 
+    # Don’t use @lru_cache(maxsize=None) here, that has a high risk of
+    # memory leaks.  It caches forever — and it keeps strong
+    # references to all function arguments and results. If the method
+    # is on a class instance (self), and the cache calls involving
+    # self, then self gets kept alive — even if no other code
+    # references it! That is a high risk of memory leaks when
+    # instantiated class objects go out of scope.  With a bounded
+    # cache, Python will evict the oldest cache entries automatically
+    # when the cache grows beyond 500,000 entries. That is much safer
+    # The self reference can still stay around as long as there are
+    # still entries in the cache for that instance. But if all entries
+    # referring to a self are evicted, then self can be garbage
+    # collected properly.
+    @functools.lru_cache(maxsize=500_000)
     def spellcheck_suggest(self, word: str) -> List[str]:
         '''Return spellchecking suggestions for word using enchant,
         pyhunspell or voikko
@@ -400,7 +403,7 @@ class Hunspell:
             else:
                 LOGGER.debug(
                     'Hunspell.init_dictionaries() dictionary_names=()\n')
-        self._suggest_cache = {}
+        self.suggest.cache_clear()
         self._dictionaries = []
         for dictionary_name in self._dictionary_names:
             self._dictionaries.append(Dictionary(name=dictionary_name))
@@ -590,6 +593,20 @@ class Hunspell:
                 dictionary_names.append(dictionary.name)
         return sorted(dictionary_names)
 
+    # Don’t use @lru_cache(maxsize=None) here, that has a high risk of
+    # memory leaks.  It caches forever — and it keeps strong
+    # references to all function arguments and results. If the method
+    # is on a class instance (self), and the cache calls involving
+    # self, then self gets kept alive — even if no other code
+    # references it! That is a high risk of memory leaks when
+    # instantiated class objects go out of scope.  With a bounded
+    # cache, Python will evict the oldest cache entries automatically
+    # when the cache grows beyond 500,000 entries. That is much safer
+    # The self reference can still stay around as long as there are
+    # still entries in the cache for that instance. But if all entries
+    # referring to a self are evicted, then self can be garbage
+    # collected properly.
+    @functools.lru_cache(maxsize=500_000)
     def suggest(self, input_phrase: str) -> List[Tuple[str, int]]:
         # pylint: disable=line-too-long
         '''Return completions or corrections for the input phrase
@@ -655,10 +672,17 @@ class Hunspell:
         >>> h = Hunspell(['None', 'en_US'])
         >>> ('Camelot', 0) in h.suggest('camel') # Should work with aspell and hunspell
         True
+
+        >>> h = Hunspell(['en_US'])
+        >>> all((x, 0) in h.suggest('tree') for x in ('trees', 'treed', "tree's"))
+        True
+
+        For benchmarking:
+
+        >>> result = [h.suggest(c) for c in 'abcdefghijklmnopqrstuvwxyz']
+        >>> result = [h.suggest(word) for word in ('infl', 'mess', 'tree', 'temp', 'wood', 'gard', 'came', 'wind', 'wate', 'succ', 'dict', 'load', 'enco', 'spell', 'word', 'auto', 'mode', 'bicy', 'moun', 'batt', 'tabl', 'star', 'moon', 'univ')]
         '''
         # pylint: enable=line-too-long
-        if input_phrase in self._suggest_cache:
-            return self._suggest_cache[input_phrase]
         if DEBUG_LEVEL > 1:
             LOGGER.debug(
                 "Hunspell.suggest() input_phrase=%(ip)s\n",
@@ -674,15 +698,16 @@ class Hunspell:
         # I.e. if '/' is already contained in the input, it cannot
         # match a word in the dictionary and we return an empty list
         # immediately:
-        if '/' in input_phrase:
-            self._suggest_cache[input_phrase] = []
+        if not input_phrase or '/' in input_phrase:
             return []
         # make sure input_phrase is in the internal normalization form (NFD):
         input_phrase = unicodedata.normalize(
             itb_util.NORMALIZATION_FORM_INTERNAL, input_phrase)
 
-        suggested_words: Dict[str, int] = {}
+        suggested_words: Dict[str, Dict[str, int]] = {}
         for dictionary in self._dictionaries:
+            name = dictionary.name
+            suggested_words[name] = {}
             if dictionary.words:
                 if dictionary.word_pairs:
                     input_phrase_no_accents = itb_util.remove_accents(
@@ -693,39 +718,15 @@ class Hunspell:
                 # complete it, it just wastes time then.
                 if len(input_phrase) <= dictionary.max_word_len:
                     if dictionary.word_pairs:
-                        if IMPORT_REGEX_SUCCESFUL:
-                            regex_pattern = regex.compile(
-                                regex.escape(input_phrase_no_accents),
-                                regex.IGNORECASE)
-                            suggested_words.update([
-                                (x[0], 0)
-                                for x in dictionary.word_pairs
-                                if regex_pattern.match(x[1])])
-                        else:
-                            re_pattern = re.compile( # pylint: disable=used-before-assignment
-                                re.escape(input_phrase_no_accents),
-                                re.IGNORECASE)
-                            suggested_words.update([
-                                (x[0], 0)
-                                for x in dictionary.word_pairs
-                                if re_pattern.match(x[1])])
+                        query = input_phrase_no_accents.casefold()
+                        for word, stripped in dictionary.word_pairs:
+                            if stripped.casefold().startswith(query):
+                                suggested_words[name][word] = 0
                     else:
-                        if IMPORT_REGEX_SUCCESFUL:
-                            regex_pattern = regex.compile(
-                                regex.escape(input_phrase),
-                                regex.IGNORECASE)
-                            suggested_words.update([
-                                (x, 0)
-                                for x in dictionary.words
-                                if regex_pattern.match(x)])
-                        else:
-                            re_pattern = re.compile(
-                                re.escape(input_phrase),
-                                re.IGNORECASE)
-                            suggested_words.update([
-                                (x, 0)
-                                for x in dictionary.words
-                                if re_pattern.match(x)])
+                        query = input_phrase.casefold()
+                        for word in dictionary.words:
+                            if word.casefold().startswith(query):
+                                suggested_words[name][word] = 0
                 if len(input_phrase) >= 4:
                     if dictionary.spellcheck(input_phrase):
                         # This is a valid word in this dictionary.
@@ -737,15 +738,24 @@ class Hunspell:
                         # knows about this, if the spell checker
                         # thinks it is a correct word, it must be
                         # counted as a match of course:
-                        suggested_words[input_phrase] = 0
-                    extra_suggestions = [
+                        suggested_words[name][input_phrase] = 0
+                    # Add suffixes:
+                    # See: https://github.com/mike-fabian/ibus-typing-booster/issues/799
+                    suffixed_suggestions = {
+                        x: 0
+                        for suggestion in sorted(suggested_words[name])[0:4]
+                        for x in dictionary.spellcheck_suggest(suggestion)
+                        if x.startswith(suggestion)
+                    }
+                    suggested_words[name].update(suffixed_suggestions)
+                    extra_suggestions = (
                         unicodedata.normalize(
                             itb_util.NORMALIZATION_FORM_INTERNAL, x)
                         for x in
                         dictionary.spellcheck_suggest(input_phrase)
-                    ]
+                    )
                     for index, suggestion in enumerate(extra_suggestions):
-                        if suggestion not in suggested_words:
+                        if suggestion not in suggested_words[name]:
                             if (dictionary.word_pairs
                                 and
                                 itb_util.remove_accents(
@@ -753,17 +763,18 @@ class Hunspell:
                                     keep=itb_util.ACCENT_LANGUAGES[
                                         dictionary.language])
                                 == input_phrase_no_accents):
-                                suggested_words[suggestion] = 0
+                                suggested_words[name][suggestion] = 0
                             else:
-                                suggested_words[suggestion] = -(index + 1)
+                                suggested_words[name][suggestion] = -(index + 1)
+        suggested_words_total = itb_util.merge_dicts_max(
+            *suggested_words.values())
         sorted_suggestions = sorted(
-            suggested_words.items(),
+            suggested_words_total.items(),
             key=lambda x: (
                 - x[1],    # 0: in dictionary, negative: spellcheck
                 len(x[0]), # length of word ascending
                 x[0],      # alphabetical
             ))[0:MAX_WORDS]
-        self._suggest_cache[input_phrase] = sorted_suggestions
         return sorted_suggestions
 
 BENCHMARK = True
